/* KallistiGL for KOS 1.1.3

   gltrans.c
   (c)2001 Dan Potter
*/

static char id[] = "KGL $Id: gltrans.c,v 1.1 2001/09/09 23:58:37 bard Exp $";

#include <GL/gl.h>
#include <GL/glu.h>
#include "glinternal.h"

#include <math.h>
#include <string.h>
#include <stdio.h>
#include <dc/fmath.h>
#include <dcutils/3dutils.h>

/* Transformation APIs */

/* Modus operandi, as it were

The transformation API will store N matrices, representing the various
GL matrix modes (projection, modelview, texture) and gl_matrix_mode will
point to the currently selected on. All matrix operations will load the matrix
from its array position, perform the math on that matrix, and store it back
into the array.

When gldraw() uses the matrices, it will pull the matrices it needs and
multiply them out in registers before applying them to vectors.

*/

/* Matrix stack (move to global?) */
#define GL_MAT_STACK_CNT 8
static matrix_t gl_mat_stack[GL_MAT_STACK_CNT];
static int gl_mat_stack_top = 0;

/* Set which matrix we are working on */
void glMatrixMode(GLenum mode) {
	gl_matrix_mode = mode;
	gl_matrix_dirty = GL_TRUE;
}

/* Load the identitiy matrix */
void glLoadIdentity(void) {
	mat_identity();
	mat_store(gl_trans_mats + gl_matrix_mode);
	gl_matrix_dirty = GL_TRUE;
}

void glOrtho(GLdouble left, GLdouble right,
		GLdouble bottom, GLdouble top,
		GLdouble near_val, GLdouble far_val) {
	gl_matrix_dirty = GL_TRUE;
}

/* Reset the GL viewport */
void glViewPort(GLint x, GLint y,
		GLsizei width, GLsizei height) {
	gl_viewport_x1 = x;
	gl_viewport_y1 = y;
	gl_viewport_width = width;
	gl_viewport_height = height;
	gl_matrix_dirty = GL_TRUE;
}

/* Located here for now */
void gluPerspective(float angle, float aspect, float znear, float zfar) {
	matrix_t screen_matrix = {
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f }
	};
	float xc, yc;

	matrix_t proj_matrix = {
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f }
	};
	float cot_fovy_2;

	/* First, the screen matrix */
	xc = gl_viewport_width / 2.0f;
	yc = gl_viewport_height / 2.0f;
	screen_matrix[0][0] = yc;
	screen_matrix[1][1] = yc;
	screen_matrix[2][2] = 1.0f;
	screen_matrix[3][0] = xc;
	screen_matrix[3][1] = yc;
	screen_matrix[3][3] = 1.0f;

	mat_apply(&screen_matrix);

	/* Second, the perspective projection matrix */
	cot_fovy_2 = 1.0f / tan(angle * 2 * F_PI / 360.0f);
	proj_matrix[0][0] = cot_fovy_2;
	proj_matrix[1][1] = -cot_fovy_2;
	proj_matrix[2][2] = (zfar + znear) / (znear - zfar);
	proj_matrix[2][3] = -1.0f;
	proj_matrix[3][2] = 2.0f * zfar * znear / (znear - zfar);
	proj_matrix[3][3] = 1.0f;
	
	mat_apply(&proj_matrix);
	mat_store(gl_trans_mats + gl_matrix_mode);

	gl_matrix_dirty = GL_TRUE;
}

void glPushMatrix(void) {
	if (gl_mat_stack_top >= GL_MAT_STACK_CNT) {
		dbglog(DBG_DEBUG, "gl: glPushMatrix() -- stack overflow\n");
		return;
	}
	
	memcpy(gl_mat_stack + gl_mat_stack_top,
		gl_trans_mats + gl_matrix_mode,
		sizeof(matrix_t));
	gl_mat_stack_top++;
}

void glPopMatrix(void) {
	if (gl_mat_stack_top <= 0) {
		dbglog(DBG_DEBUG, "gl: glPopMatrix() -- stack underflow\n");
		return;
	}

	gl_mat_stack_top--;
	memcpy(gl_trans_mats + gl_matrix_mode,
		gl_mat_stack + gl_mat_stack_top,
		sizeof(matrix_t));

	gl_matrix_dirty = GL_TRUE;
}

void glRotatef(GLfloat angle,
		GLfloat x, GLfloat y, GLfloat z) {
	matrix_t mx = {
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f }
	}, my = {
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f }
	}, mz = {
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f }
	};
	float c, s;

	mat_load(gl_trans_mats + gl_matrix_mode);

	/* Rotation about X axis */
	c = fcos(angle * x);
	s = fsin(angle * x);
	mx[0][0] = 1.0f;
	mx[1][1] = c;
	mx[1][2] = -s;
	mx[2][1] = s;
	mx[2][2] = c;
	mx[3][3] = 1.0f;
	mat_apply(&mx);

	/* Rotation about Y axis */
	c = fcos(angle * y);
	s = fsin(angle * y);
	my[0][0] = c;
	my[0][2] = s;
	my[1][1] = 1.0f;
	my[2][0] = -s;
	my[2][2] = c;
	my[3][3] = 1.0f;
	mat_apply(&my);

	/* Rotation about Z axis */
	c = fcos(angle * z);
	s = fsin(angle * z);
	mz[0][0] = c;
	mz[0][1] = -s;
	mz[1][0] = s;
	mz[1][1] = c;
	mz[2][2] = 1.0f;
	mz[3][3] = 1.0f;
	mat_apply(&mz);

	mat_store(gl_trans_mats + gl_matrix_mode);

	gl_matrix_dirty = GL_TRUE;
}

void glScalef(GLfloat x, GLfloat y, GLfloat z) {
	matrix_t m = {
		{    x, 0.0f, 0.0f, 0.0f },
		{ 0.0f,    y, 0.0f, 0.0f },
		{ 0.0f, 0.0f,    z, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 1.0f }
	};
	mat_load(gl_trans_mats + gl_matrix_mode);
	mat_apply(&m);
	mat_store(gl_trans_mats + gl_matrix_mode);

	gl_matrix_dirty = GL_TRUE;
}

void glTranslatef(GLfloat x, GLfloat y, GLfloat z) {
	matrix_t m = {
		{ 1.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 1.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 1.0f, 0.0f },
		{    x,    y,    z, 1.0f }
	};
	mat_load(gl_trans_mats + gl_matrix_mode);
	mat_apply(&m);
	mat_store(gl_trans_mats + gl_matrix_mode);

	gl_matrix_dirty = GL_TRUE;
}


